/*
 * Connection tracking state replication for Netfilter
 *
 * (C) 2002-2003, KOVACS Krisztian <hidden@sch.bme.hu>
 * (C) 2003-2004, Harald Welte <laforge@netfilter.org>
 * (C) 2004, BalaBit IT Ltd. <www.balabit.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * $Id: ct_sync_main.c,v 1.39 2004/08/30 11:05:46 hidden Exp $
 *
 */

#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/prefetch.h>
#include <linux/proc_fs.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/skbuff.h>
#include <linux/notifier.h>
#include <linux/sysctl.h>
#include <net/sock.h>

#include <linux/netdevice.h>
#include <linux/netlink.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_packet.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>
#include <linux/netfilter_ipv4/ip_nat_core.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>

#include <asm/uaccess.h>
#include <linux/unistd.h>
static int errno; /* FIXME: linux/unistd.h needs this */

#include <linux/netfilter_ipv4/ct_sync.h>
#include <linux/netfilter_ipv4/ct_sync_proto.h>

#define DEBUGP	printk

#if (CTSYNC_DEBUG > 0)
#define CT_SYNC_DUMP_TUPLE(x) DUMP_TUPLE(x)
#else
#define CT_SYNC_DUMP_TUPLE(x)
#endif

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_conntrack_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_conntrack_lock)
#include <linux/netfilter_ipv4/listhelp.h>

#define CT_SYNC_VERSION	"0.16"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("KOVACS Krisztian <hidden@sch.bme.hu>, Harald Welte <laforge@netfilter.org>");
MODULE_DESCRIPTION("Connection tracking state synchronization module");

/* module parameters */
static char *syncdev = "";
static int state = -1;
static int id = -1;
static int l2drop = 0;
static int cmarkbit = sizeof(unsigned long) * 8 - 1;

MODULE_PARM(syncdev, "s");
MODULE_PARM_DESC(syncdev, "Interface used for sync messages");
MODULE_PARM(state, "i");
MODULE_PARM_DESC(state, "Initial state: 0=slave 1=master");
MODULE_PARM(id, "i");
MODULE_PARM_DESC(id, "unique Node ID (0..255) of local node");
MODULE_PARM(l2drop, "i");
MODULE_PARM_DESC(l2drop, "drop all packets at layer 2 if slave");
MODULE_PARM(cmarkbit, "i");
MODULE_PARM_DESC(cmarkbit, "bit in the connection mark to use as sync mark");

/* used to stop ct_sync threads */
static DECLARE_WAIT_QUEUE_HEAD(stop_ct_rcv_thread_wait);
static atomic_t stop_ct_rcv_thread = ATOMIC_INIT(0);

static DECLARE_WAIT_QUEUE_HEAD(ct_sync_send_wait);
static DECLARE_WAIT_QUEUE_HEAD(stop_ct_send_thread_wait);
static atomic_t stop_ct_send_thread = ATOMIC_INIT(0);

static atomic_t ct_sync_initsync_active = ATOMIC_INIT(0);
static DECLARE_WAIT_QUEUE_HEAD(ct_sync_initsync_wait);
static DECLARE_WAIT_QUEUE_HEAD(stop_ct_initsync_thread_wait);
static atomic_t stop_ct_initsync_thread = ATOMIC_INIT(0);

/* configuration data */
static struct ct_sync_cfg
{
	struct sockaddr_in addr;	/* multicast address */
	char devname[IFNAMSIZ];		/* dedicated sync network device */
	int ifindex;			/* ifindex of the network device */
	struct cts_protoh *protoh;	/* protocol handle */
	u8 nodeid;			/* node id in cluster */
} cts_cfg;

/***********************************************************************
 * FILLING CTSYNC MESSAGES WITH DATA
 ***********************************************************************/

static int
fill_expectmsg(void *buff, __u8 event,
	       struct ip_conntrack *master,
	       struct ip_conntrack_expect *expect)
{
	struct ct_sync_msghdr *hdr = buff;
	struct ct_sync_expect *sexp = buff + sizeof(*hdr);

	CT_SYNC_ENTER();

	/* fill event header */
	hdr->type = event;
	hdr->resource = CT_SYNC_RES_EXPECT;
	hdr->len = __constant_htons(sizeof(*sexp));

	/* copy data from expectation */
	memcpy(&sexp->tuple, &expect->tuple, sizeof(sexp->tuple));
	memcpy(&sexp->mask, &expect->mask, sizeof(sexp->mask));
	sexp->seq = expect->seq;
	memcpy(&sexp->proto, &expect->proto, sizeof(sexp->proto));
	sexp->help = expect->help;
	if (expect->expectant)
		memcpy(&sexp->expectant,
			&master->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
			sizeof(sexp->expectant));
	else
		memset(&sexp->expectant, 0, sizeof(sexp->expectant));

	if (expect->sibling)
		memcpy(&sexp->sibling,
			&expect->sibling->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
			sizeof(sexp->sibling));
	else
		memset(&sexp->sibling, 0, sizeof(sexp->sibling));

	CT_SYNC_LEAVE();

	return 0;
}

static int
fill_ctmsg(void *buff, __u8 event, struct ip_conntrack *ct)
{
	struct ct_sync_msghdr *hdr = buff;
	struct ct_sync_conntrack *sct = buff + sizeof(*hdr);
#ifdef CONFIG_IP_NF_TARGET_MASQUERADE
	struct net_device *masq_dev;
#endif

	CT_SYNC_ENTER();

	/* fill event header */
	hdr->type = event;
	hdr->resource = CT_SYNC_RES_CONNTRACK;
	hdr->len = __constant_htons(sizeof(*sct));

	CT_SYNC_DUMP_TUPLE(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple);
	CT_SYNC_DUMP_TUPLE(&ct->tuplehash[IP_CT_DIR_REPLY].tuple);

	/* copy data from conntrack struct */
	memset(sct, 0, sizeof(*sct));
	memcpy(&sct->orig, &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple, 
		sizeof(sct->orig));
	memcpy(&sct->reply, &ct->tuplehash[IP_CT_DIR_REPLY].tuple,
		sizeof(sct->reply));
	sct->status = (u32) ct->status;
	sct->expires = ct->timeout.expires - jiffies;
	if (unlikely(ct->helper != NULL)) {
		strncpy(sct->helper, ct->helper->name, sizeof(sct->helper));
		memcpy(&sct->help, &ct->help, sizeof(sct->help));
	} else
		sct->helper[0] = '\0';

	sct->expecting = ct->expecting;

	if (unlikely(ct->master != NULL)) {
		struct ip_conntrack *master_ct = master_ct(ct);
		CT_SYNC_ASSERT(master_ct);
		if (master_ct)
			memcpy(&sct->master_ct_tuple,
				&master_ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
				sizeof(sct->master_ct_tuple));
		else
			memset(&sct->master_ct_tuple, 0, sizeof(sct->master_ct_tuple));
	}

	memcpy(&sct->proto, &ct->proto, sizeof(sct->proto));

#ifdef CONFIG_IP_NF_NAT_NEEDED
	//if (likely(ct->nat.info.initialized && ct->nat.info.num_manips)) {
	if (likely(ct->nat.info.initialized)) {
		const struct ip_nat_info *nat = &ct->nat.info;

		sct->nat_initialized = nat->initialized;
		sct->nat_num_manips = nat->num_manips;
		memcpy(sct->nat_manips, &nat->manips, 
			(nat->num_manips * sizeof(struct ip_nat_info_manip)));
		if (unlikely(nat->helper != NULL)) {
			strncpy(sct->nat_helper, nat->helper->name, 
				sizeof(sct->nat_helper));
			memcpy(&sct->nat_help, &ct->nat.help, 
				sizeof(sct->nat_help));
		} else
			sct->nat_helper[0] = '\0';

		memcpy(&sct->nat_seq, &nat->seq, sizeof(sct->nat_seq));
#ifdef CONFIG_IP_NF_TARGET_MASQUERADE
		if ((masq_dev = dev_get_by_index(ct->nat.masq_index)) != NULL) {
			memcpy(&sct->nat_masq_iface, &masq_dev->name, sizeof(masq_dev->name));
			dev_put(masq_dev);
		} else {
			sct->nat_masq_iface[0] = '\0';
		}
#endif
	} else
		sct->nat_initialized = 0;
#endif /* CONFIG_IP_NF_NAT_NEEDED */

#ifdef CONFIG_IP_NF_CONNTRACK_MARK
	sct->mark = (u32) ct->mark;
#endif

	CT_SYNC_LEAVE();

	return 0;
}

/***********************************************************************
 * BACKEND FOR CONNTRACK MODIFICATION
 ***********************************************************************/

static inline int
_helper_name_cmp(const struct ip_conntrack_helper *i,
		 const char *name)
{
	return strncmp(i->name, name, CT_SYNC_CTHELPERSIZE - 1);
}

static inline int
_nat_helper_name_cmp(const struct ip_nat_helper *i, const char *name)
{
	return strncmp(i->name, name, CT_SYNC_NATHELPERSIZE - 1);
}

/*
 * update data in the conntrack entry
 */
static inline int
_ct_sync_update_conntrack(struct ip_conntrack *ct,
			  struct ct_sync_conntrack *sct, int new)
{
	struct ip_conntrack_helper *helper;
	int previous_nat_initialized = 0;
#ifdef CONFIG_IP_NF_TARGET_MASQUERADE
	struct net_device *masq_dev;
#endif

	CT_SYNC_ENTER();

	/* fill tuples for new conntracks, tuples must not be changed
	 * for confirmed connections */
	if (new) {
		memcpy(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple, &sct->orig,
		       sizeof(struct ip_conntrack_tuple));
		memcpy(&ct->tuplehash[IP_CT_DIR_REPLY].tuple, &sct->reply,
		       sizeof(struct ip_conntrack_tuple));
	} else {
		CT_SYNC_ASSERT(memcmp(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
				    &sct->orig, sizeof(sct->orig)) == 0);
		CT_SYNC_ASSERT(memcmp(&ct->tuplehash[IP_CT_DIR_REPLY].tuple,
				    &sct->reply, sizeof(sct->reply)) == 0);
		CT_SYNC_DUMP_TUPLE(&sct->orig);
		CT_SYNC_DUMP_TUPLE(&sct->reply);
	}

	CT_SYNC_DUMP_TUPLE(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple);
	CT_SYNC_DUMP_TUPLE(&ct->tuplehash[IP_CT_DIR_REPLY].tuple);

	/* optimization: done by conntrack_alloc */
	//ct->tuplehash[IP_CT_DIR_ORIGINAL].ctrack = ct;
	//ct->tuplehash[IP_CT_DIR_REPLY].ctrack = ct;

	/* timeout */
	if (!timer_pending(&ct->timeout)) {
		ct->timeout.expires = sct->expires;
	} else {
		CT_SYNC_DEBUG("Trying to update an active conntrack entry!\n");
		CT_SYNC_LEAVE();
		return -1;
	}

	/* status */
	ct->status = sct->status;

	/* sibling_list initialized by conntrack_alloc */

	/* FIXME: sibling expectations */
	//ct->expecting = sct->expecting;
	ct->expecting = 0;

	/* FIXME: have to look up the master_ct by tuple and then iterate
	 * over ip_conntrack_expect's in order to link with the correct
	 * expect */
#if 0
	if (sct->exp_master) {
		struct ip_conntrack_expect *expect;
		idmap = ct_sync_idmap_find(sct->exp_master);
		if (!idmap) {
			CT_SYNC_ERR("Unknown master expectation, ignoring.\n");
		} else {
			expect = idmap->ptr;
			ct->master = expect;
			expect->sibling = ct;
			/* FIXME: delete expect from expect list and decrease expectant's count */
		}
	}
#endif

	/* conntrack helper */
	if (unlikely(strcmp(sct->helper, ""))) {
		READ_LOCK(&ip_conntrack_lock);
		helper = LIST_FIND(&ip_conntrack_helpers, 
				   _helper_name_cmp,
				   struct ip_conntrack_helper *,
				   sct->helper);
		if (unlikely(!helper)) {
			CT_SYNC_ERR("Unknown helper, ignoring.\n");
			ct->helper = NULL;
		} else {
			ct->helper = helper;
			memcpy(&ct->help, &sct->help, sizeof(ct->help));
		}
		READ_UNLOCK(&ip_conntrack_lock);
	}

	/* nf_ct_infos filled in by conntrack_alloc */
	
	/* protocol data */
	memcpy(&ct->proto, &sct->proto, sizeof(ct->proto));

	/* NAT */
#ifdef CONFIG_IP_NF_NAT_NEEDED
	//if (likely(sct->nat_initialized && sct->nat_num_manips)) {
	if (likely(sct->nat_initialized )) {
		struct ip_nat_info *nat = &ct->nat.info;
		previous_nat_initialized = nat->initialized;

		/* make sure we don't modify the NAT info of an already
		 * confirmed entry */
		CT_SYNC_ASSERT(new || (nat->initialized == sct->nat_initialized));
		CT_SYNC_DEBUG("NAT initialized: %x, new value %x\n", nat->initialized,
				sct->nat_initialized);

		if (new) {
			nat->initialized = sct->nat_initialized;
			nat->bysource.conntrack = ct;
			nat->byipsproto.conntrack = ct;
			nat->num_manips = sct->nat_num_manips;
			memcpy(&nat->manips, sct->nat_manips, 
				(sct->nat_num_manips * sizeof(struct ip_nat_info_manip)));

			if (unlikely(strcmp(sct->nat_helper, ""))) {
				struct ip_nat_helper *helper;
				/* look up nat helper */
				READ_LOCK(&ip_nat_lock);
				helper = LIST_FIND(&ip_nat_helpers, _nat_helper_name_cmp,
						   struct ip_nat_helper *,
						   sct->nat_helper);
				if (unlikely(!helper)) {
					CT_SYNC_ERR("Unknown NAT helper, ignoring\n");
					nat->helper = NULL;
					memset(&ct->nat.help, 0, sizeof(ct->nat.help));
				} else {
					nat->helper = helper;
					memcpy(&ct->nat.help, &sct->nat_help,
						sizeof(ct->nat.help));
				}
				READ_UNLOCK(&ip_nat_lock);
				memcpy(&nat->seq, &sct->nat_seq, sizeof(nat->seq));
			}
#ifdef CONFIG_IP_NF_TARGET_MASQUERADE
			if ((masq_dev = dev_get_by_name(sct->nat_masq_iface)) != NULL) {
				ct->nat.masq_index = masq_dev->ifindex;
				dev_put(masq_dev);
			} else {
				/* FIXME: is 0 a valid ifindex? */
				ct->nat.masq_index = 0;
			}
#endif
		}
	}
#endif /* CONFIG_IP_NF_NAT_NEEDED */

#ifdef CONFIG_IP_NF_CONNTRACK_MARK
	ct->mark = (unsigned long) sct->mark;
#endif

	/* add to hash tables */
	if (new && is_confirmed(ct)) {
		WRITE_LOCK(&ip_conntrack_lock);
		if (!__ip_conntrack_find(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple, NULL) 
		    && !__ip_conntrack_find(&ct->tuplehash[IP_CT_DIR_REPLY].tuple, NULL)) {
			u_int32_t hash, repl_hash;
			hash = hash_conntrack(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple);
			repl_hash = hash_conntrack(&ct->tuplehash[IP_CT_DIR_REPLY].tuple);
			/* place in ordered list */
			ct->id = ip_conntrack_next_id++;
			list_add_tail(&ct->olist, &ip_conntrack_ordered_list);
			/* put in conntrack hash */
                	list_prepend(&ip_conntrack_hash[hash], &ct->tuplehash[IP_CT_DIR_ORIGINAL]);
                	list_prepend(&ip_conntrack_hash[repl_hash], &ct->tuplehash[IP_CT_DIR_REPLY]);
                	atomic_inc(&ct->ct_general.use);
			/* put in NAT hashes if necessary */
			if (sct->nat_initialized) {
				CT_SYNC_ASSERT(previous_nat_initialized == 0);
				WRITE_LOCK(&ip_nat_lock);
				place_in_hashes(ct, &ct->nat.info);
				WRITE_UNLOCK(&ip_nat_lock);
			}
		} else {
			CT_SYNC_ERR("want to put conntrack in hash but is already there\n");
		}
		WRITE_UNLOCK(&ip_conntrack_lock);
	}

	CT_SYNC_LEAVE();

	return 0;
}

static inline void
__ct_sync_unexpect_related(struct ip_conntrack_expect *expect)
{
	if (unlikely(expect->sibling != NULL)) {
		CT_SYNC_ERR("Cannot delete a confirmed expectation!\n");
		return;
	}

	list_del(&expect->list);
	list_del(&expect->expected_list);

	if (likely(expect->expectant != NULL))
		expect->expectant->expecting--;

	ip_conntrack_expect_put(expect);
}

static inline void
_ct_sync_unexpect_related(struct ip_conntrack_expect *expect)
{
	if (unlikely(!expect->expectant || !expect->expectant->helper))
		return;
	if (unlikely(expect->expectant->helper->timeout 
		     && !del_timer(&expect->timeout)))
		return;
	
	__ct_sync_unexpect_related(expect);
}

static inline void
_ct_sync_conntrack_remove_expectations(struct ip_conntrack *ct)
{
	struct list_head *exp_entry, *next;
	struct ip_conntrack_expect *exp;

	CT_SYNC_ENTER();
	
        for (exp_entry = ct->sibling_list.next;
             exp_entry != &ct->sibling_list; exp_entry = next) {
                next = exp_entry->next;
                exp = list_entry(exp_entry, struct ip_conntrack_expect,
                                 expected_list);

                if (exp->sibling)
                        continue;

                _ct_sync_unexpect_related(exp);
        }

	CT_SYNC_LEAVE();
}

static inline void
_ct_sync_conntrack_clean_lists(struct ip_conntrack *ct)
{
	u_int32_t hash, repl_hash;

	CT_SYNC_ENTER();

	WRITE_LOCK(&ip_conntrack_lock);

	hash = hash_conntrack(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple);
	repl_hash = hash_conntrack(&ct->tuplehash[IP_CT_DIR_REPLY].tuple);
	LIST_DELETE(&ip_conntrack_hash[hash], 
		    &ct->tuplehash[IP_CT_DIR_ORIGINAL]);
	LIST_DELETE(&ip_conntrack_hash[repl_hash],
		    &ct->tuplehash[IP_CT_DIR_REPLY]);
	LIST_DELETE(&ip_conntrack_ordered_list, &ct->olist);

	WRITE_UNLOCK(&ip_conntrack_lock);

	_ct_sync_conntrack_remove_expectations(ct);

	CT_SYNC_LEAVE();
}

/* delete conntrack entry */
static inline int
_ct_sync_remove_conntrack(struct ip_conntrack *ct)
{
	CT_SYNC_ENTER();

	del_timer(&ct->timeout);
	_ct_sync_conntrack_clean_lists(ct);
	ip_conntrack_put(ct);

	CT_SYNC_LEAVE();

	return 0;
}

static inline int
_start_ct_timer(const struct ip_conntrack_tuple_hash *h,
		unsigned int *num, unsigned int *started)
{
	struct ip_conntrack *ct = h->ctrack;

	if (DIRECTION(h))
		return 0;

	(*num)++;
	if (likely(is_confirmed(ct))) {
		if (!timer_pending(&ct->timeout)) {
			/* add five extra seconds to make sure that the entry
			 * won't timeout instantaneously */
			ct->timeout.expires += jiffies + 5 * HZ;
			add_timer(&ct->timeout);
			(*started)++;
		}
	}

	return 0;
}

static void
ct_sync_start_timers(void)
{
	int i;
	unsigned int num_started = 0;
	unsigned int num_entries = 0;

	CT_SYNC_ENTER();

	WRITE_LOCK(&ip_conntrack_lock);

	for (i = 0; i < ip_conntrack_htable_size; i++) {
		if (LIST_FIND(&ip_conntrack_hash[i], _start_ct_timer,
			      struct ip_conntrack_tuple_hash *, 
			      &num_entries, &num_started))
			break;
	}

	WRITE_UNLOCK(&ip_conntrack_lock);

	CT_SYNC_INFO("started timer of %u (total %u) conntrack entries\n", 
		     num_started, num_entries);

	CT_SYNC_LEAVE();
}

static inline int
_stop_ct_timer(const struct ip_conntrack_tuple_hash *h, 
		unsigned int *num, unsigned int *stopped)
{
	struct ip_conntrack *ct = h->ctrack;

	if (DIRECTION(h))
		return 0;

	(*num)++;
	if (is_confirmed(ct)
#ifdef CONFIG_IP_NF_CONNTRACK_MARK
	    /* stop timer only if this is a synchronized connection */
	    && test_bit(cmarkbit, &ct->mark)
#endif
	) {
		if (del_timer(&ct->timeout))
			(*stopped)++;
	}

	return 0;
}

static void
ct_sync_stop_timers(void)
{
	int i;
	unsigned int num_entries = 0;
	unsigned int num_stopped = 0;

	CT_SYNC_ENTER();

	WRITE_LOCK(&ip_conntrack_lock);

	for (i = 0; i < ip_conntrack_htable_size; i++) {
		if (LIST_FIND(&ip_conntrack_hash[i], _stop_ct_timer,
			      struct ip_conntrack_tuple_hash *,
			      &num_entries, &num_stopped))
			break;
	}

	WRITE_UNLOCK(&ip_conntrack_lock);

	CT_SYNC_INFO("stopped timer of %u (total %u) conntrack entries\n", 
			num_stopped, num_entries);

	CT_SYNC_LEAVE();
}

static int kill_all(const struct ip_conntrack *i, void *data)
{
	return 1;
}


/***********************************************************************
 * MESSAGE PROCESSING FUNCTIONS
 ***********************************************************************/

/* process messages */
static int
ct_sync_msg_process_updateconntrack(void *data, u16 len)
{
	struct ct_sync_conntrack *sct = (struct ct_sync_conntrack *)data;
	struct ip_conntrack_tuple_hash *h;
	struct ip_conntrack *ct;
	struct ip_conntrack_tuple dummy_tuple;
	int new = 0;

	CT_SYNC_ENTER();

	if (unlikely(len < sizeof(struct ct_sync_conntrack))) {
		CT_SYNC_ERR("Illegal message size: %u\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}

	h = ip_conntrack_find_get(&sct->orig, NULL);
	if (h) {
		ct = h->ctrack;
	} else {
		CT_SYNC_DEBUG("Conntrack entry not found, creating.\n");
		ct = ip_conntrack_alloc(&dummy_tuple, &dummy_tuple);
		if (unlikely(ct == NULL)) {
			CT_SYNC_DEBUG("Unable to create conntrack entry, ignoring message.\n");
			CT_SYNC_LEAVE();
			return -1;
		}
		new = 1;
	}

	_ct_sync_update_conntrack(ct, sct, new);

	ip_conntrack_put(ct);

	CT_SYNC_DEBUG("updated conntrack: ");
	CT_SYNC_DUMP_TUPLE(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple);

	CT_SYNC_LEAVE();

	return 0;
}

static int
ct_sync_msg_process_delconntrack(void *data, u16 len)
{
	struct ip_conntrack_tuple *t = (struct ip_conntrack_tuple * ) data;
	struct ip_conntrack_tuple_hash *h;
		
	CT_SYNC_ENTER();

	if (unlikely(len < sizeof(*t))) {
		CT_SYNC_ERR("Illegal message size: %u\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}

	h = ip_conntrack_find_get(t, NULL);
	if (h) {
		_ct_sync_remove_conntrack(h->ctrack);
		CT_SYNC_DEBUG("Deleting conntrack: ");
		CT_SYNC_DUMP_TUPLE(t);
		ip_conntrack_put(h->ctrack);
	} else {
		CT_SYNC_DEBUG("Cannot delete nonexistent conntrack:");
		CT_SYNC_DUMP_TUPLE(t);
	}

	CT_SYNC_LEAVE();

	return 0;
}

static int
ct_sync_msg_process_updateexpect(void *data, u16 len)
{
	struct ct_sync_expect *exp = (struct ct_sync_expect *)data;

	CT_SYNC_ENTER();

	if (unlikely(len < sizeof(struct ct_sync_expect))) {
		CT_SYNC_ERR("Illegal message size: %u\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}

	CT_SYNC_LEAVE();

	return 0;
}

static int
ct_sync_msg_process_delexpect(void *data, u16 len)
{
	struct ct_sync_expect *sexp = (struct ct_sync_expect *)data;
	struct ip_conntrack_expect *exp;

	CT_SYNC_ENTER();

	if (unlikely(len < sizeof(struct ct_sync_expect))) {
		CT_SYNC_ERR("Illegal message size: %u\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}
#if 0
	READ_LOCK(&ip_conntrack_lock);
	WRITE_LOCK(&ip_conntrack_expect_tuple_lock);
	exp = LIST_FIND(&ip_conntrack_expect_list, expect_cmp,
			struct ip_conntrack_expect *, exp->tuple);
	if (!exp || !exp->expectant) 
		goto unlock_out;

	if (conntrack_tuple_cmp(&exp->expectant,
			&exp->expectant->tuplehash[IP_CT_DIR_ORIGINAL].tuple))
		__unexpect_related(exp);

unlock_out:
	WRITE_INLOCK(&ip_conntrack_expect_tuple_lock);
	READ_UNLOCK(&ip_donntrack_lock);
#endif
	CT_SYNC_LEAVE();
	return 0;
}

/* process incoming message in slave mode */
static int
ct_sync_msg_process_slave(struct ct_sync_msghdr *msghdr, 
			  struct ct_sync_pkthdr *pkthdr)
{
	int ret = -1;
	char *data = (char *)msghdr + sizeof(*msghdr);
	int len = ntohs(msghdr->len);

	CT_SYNC_ENTER();

	if (pkthdr->pkttype != CT_SYNC_PKT_SYNC) {
		CT_SYNC_DEBUG("Invalid packet while being SLAVE: %d\n",
				pkthdr->pkttype);
		CT_SYNC_LEAVE();
		return -1;
	}

	switch (msghdr->resource) {
	case CT_SYNC_RES_CONNTRACK:
		switch (msghdr->type) {
		case CT_SYNC_MSG_UPDATE:
			ret = ct_sync_msg_process_updateconntrack(data, len);
			break;
		case CT_SYNC_MSG_DELETE:
			ret = ct_sync_msg_process_delconntrack(data, len);
			break;
		}
		break;
	case CT_SYNC_RES_EXPECT:
		switch(msghdr->type) {
		case CT_SYNC_MSG_UPDATE:
			ret = ct_sync_msg_process_updateexpect(data, len);
			break;
		case CT_SYNC_MSG_DELETE:
			ret = ct_sync_msg_process_delexpect(data, len);
			break;
		}
		break;
	default:
		CT_SYNC_DEBUG("Invalid resource while being SLAVE: %d.\n",
				msghdr->resource); 
	}

	CT_SYNC_LEAVE();

	return ret;
}

/* process incoming message in master mode */
static int
ct_sync_msg_process_master(struct ct_sync_msghdr *msghdr,
			   struct ct_sync_pkthdr *pkthdr)
{
	int ret;
	u16 pktseq = ntohs(pkthdr->pktseq);
	//u16 minseq = ntohs(pkthdr->minseq);

	CT_SYNC_ENTER();

	CT_SYNC_DEBUG("seq: %u\n", pktseq);

	switch (pkthdr->pkttype) {
		case CT_SYNC_PKT_SYNC:
			CT_SYNC_DEBUG("received SYNC while master->ignoring\n");
			ret = 0;
			break;
		default:
			CT_SYNC_DEBUG("Invalid packet while being MASTER:%d\n",
					pkthdr->pkttype);
			ret = -1;
			break;
	}

	CT_SYNC_LEAVE();

	return ret;
}

/* start initial synchronization */
static int
ct_sync_start_initsync(struct cts_protoh *cph)
{
	CT_SYNC_ENTER();
	
	if (cts_proto_is_master(cts_cfg.protoh) &&
	    !atomic_read(&ct_sync_initsync_active)) {
		/* signal that an initsync is to follow */
		if (cts_proto_initsync_follows(cph) < 0) {
			CT_SYNC_LEAVE();
			return -1;
		}
		/* wake up initsync thread */
		wake_up(&ct_sync_initsync_wait);
	} else
		CT_SYNC_INFO("not master, or initsync already in progress\n");

	CT_SYNC_LEAVE();

	return 0;
}

/***********************************************************************
 * KERNEL THREADS
 ***********************************************************************/

#define CT_SYNC_SEND_BURST 	100
/* sync kernel thread: sender */
static int
ct_sync_send_thread_main(void *ct_thread_startup)
{
	mm_segment_t oldmm;
	unsigned long remaining;
	DECLARE_WAITQUEUE(wait, current);
	
	CT_SYNC_ENTER();

	/* daemonize kernel thread, set values */
	daemonize();
	oldmm = get_fs();
	set_fs(KERNEL_DS);
	sigfillset(&current->blocked);
	sprintf(current->comm, "ct_sync_send");
	
	complete((struct completion *) ct_thread_startup);

        for (;;) {
		int i = 0;
		if (unlikely(atomic_read(&stop_ct_send_thread))) {
			CT_SYNC_DEBUG("stop_ct_send_thread == 1\n");
			break;
		}

		/* if there are pending packets, dequeue and send
		 * all of them */
		while (cts_proto_send_pending(cts_cfg.protoh)) {
			cts_proto_send_dequeue(cts_cfg.protoh);
			i++;
			if (i >= CT_SYNC_SEND_BURST)
				break;
		}

		__set_current_state(TASK_INTERRUPTIBLE);
		add_wait_queue(&ct_sync_send_wait, &wait);
		CT_SYNC_DEBUG("falling asleep\n");
		remaining = schedule_timeout(HZ);
		remove_wait_queue(&ct_sync_send_wait, &wait);
		__set_current_state(TASK_RUNNING);

		if (remaining <= 0) {
			/* we had a timeout, try to advance buffer even if it's not full */
			CT_SYNC_DEBUG("woken up because schedule_timeout() timed out\n");
			if (cts_proto_advance_timeout(cts_cfg.protoh) < 0)
				CT_SYNC_ERR("advancing current csb failed\n");
		} else
			CT_SYNC_DEBUG("woken up after schedule_timeout() because we had an event\n");
        }

	set_fs(oldmm);

	atomic_set(&stop_ct_send_thread, 0);
	wake_up(&stop_ct_send_thread_wait);

	CT_SYNC_LEAVE();
	
	return 0;
}

/* FIXME: these should be configurable */
#define CT_SYNC_RECV_BURST 100
#define CT_SYNC_INITSYNC_RATE 20
/* sync kernel thread: receiver */
static int
ct_sync_rcv_thread_main(void *ct_thread_startup)
{
	mm_segment_t oldmm;
	
	CT_SYNC_ENTER();
	
	/* daemonize kernel thread, set values */
	daemonize();
	oldmm = get_fs();
	set_fs(KERNEL_DS);
	sigfillset(&current->blocked);
	sprintf(current->comm, "ct_sync_rcv");

	complete((struct completion *) ct_thread_startup);

	CT_SYNC_DEBUG("ct_sync_rcv: entering main loop\n");

	for (;;) {
		int pkts_enqueued = 0;
		int pkts_received = 0;
		int msgs_received = 0;
		struct ct_sync_msghdr *msghdr;
		struct ct_sync_pkthdr *pkthdr;

	  	if (unlikely(atomic_read(&stop_ct_rcv_thread)))
		  	break;

		/* first process all packets from the socket buffer into
		 * our protocol layer */
		while (cts_proto_recv_pending(cts_cfg.protoh)) {
			switch (cts_proto_recv_pkt(cts_cfg.protoh)) {
			case CTS_PROTO_RCV_ERROR:
				/* FIXME: unrecoverable error, should complain, but 
				 * should not exit! */
				goto error;
				break;
			case CTS_PROTO_RCV_SYNC:
				/* we'll handle this later when dequeuing messages */
				pkts_enqueued++;
				break;
			case CTS_PROTO_RCV_INITSYNC:
				/* initiate initial synchronization */
				ct_sync_start_initsync(cts_cfg.protoh);
				break;
			default:
				/* the protocol layer handled other cases, and dropped
				 * the packet */
				break;
			}
			
			pkts_received++;
			if (pkts_received >= CT_SYNC_RECV_BURST)
				break;
		}

		if (pkts_enqueued > 0)
			/* process messages */
			while ((msghdr = cts_proto_recvmsg(cts_cfg.protoh, 
								&pkthdr))) {
				msgs_received++;

				if (cts_proto_is_master(cts_cfg.protoh))
					ct_sync_msg_process_master(msghdr, 
								pkthdr);
				else
					ct_sync_msg_process_slave(msghdr, 
								pkthdr);
			}
			
		if (msgs_received < CT_SYNC_INITSYNC_RATE) {
			/* We have not received more than CT_SYNC_INITSYNC_RATE
			 * messages in the last 100 ms.  This usually means
			 * we're not receiving initial re-synchronization
			 * messages anymore, so if we're in state
			 * SLAVE_SYNRECV, we transition into SLAVE_RUNNING. */
			cts_proto_recv_running(cts_cfg.protoh);
		}

		__set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/10);
		__set_current_state(TASK_RUNNING);
	}

error:
	set_fs(oldmm);

	atomic_set(&stop_ct_rcv_thread, 0);
	wake_up(&stop_ct_rcv_thread_wait);

	CT_SYNC_LEAVE();

	return 0;
}

static inline int
_send_initsync(const struct ip_conntrack_tuple_hash *h, unsigned int *num)
{
	struct ip_conntrack *ct = h->ctrack;

	if (DIRECTION(h))
		return 0;

	if (likely(is_confirmed(ct))) {
		char *buff;
		struct cts_buff *csb;
		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb, 
				CTMSG_SIZEOF(struct ct_sync_conntrack));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to store create event\n");
			return 1;
		}
		fill_ctmsg(buff, CT_SYNC_MSG_UPDATE, ct);
		csb_use_dec(cts_cfg.protoh, csb);
		(*num)++;
	}

	return 0;
}

/* sync kernel thread: recover */
static int
ct_sync_initsync_thread_main(void *ct_thread_startup)
{
	DECLARE_WAITQUEUE(wait, current);
	mm_segment_t oldmm;

	CT_SYNC_ENTER();

	/* daemonize kernel thread, set values */
	daemonize();
	oldmm = get_fs();
	set_fs(KERNEL_DS);
	sigfillset(&current->blocked);
	sprintf(current->comm, "ct_sync_initsync");

	complete((struct completion *) ct_thread_startup);

	CT_SYNC_DEBUG("ct_sync_initsync: entering main loop\n");

	for (;;) {
		int i; 

		__set_current_state(TASK_INTERRUPTIBLE);
		add_wait_queue(&ct_sync_initsync_wait, &wait);
		schedule();
		remove_wait_queue(&ct_sync_initsync_wait, &wait);
		__set_current_state(TASK_RUNNING);

		/* somebody woke us.  either initsync or thread termination
		   request */

		if (unlikely(atomic_read(&stop_ct_initsync_thread)))
			break;

		CT_SYNC_INFO("starting initsync dump\n");

		/* speed: 40 conntracks per schedule, 100 schedules per
		   second: 4000 conntracks per second.  This makes about
		   25 seconds for 100,000 connections.  Please note that
		   we have four conntracks per packet, so the wire would
		   see something like 1000 packets per second, or in
		   other words: 1.5mbits */

		atomic_set(&ct_sync_initsync_active, 1);
		for (i = 0; i < ip_conntrack_htable_size; i++) {
			unsigned int num_sent = 0;

			if (unlikely(atomic_read(&stop_ct_initsync_thread)))
				break;

			READ_LOCK(&ip_conntrack_lock);
dump_bucket_locked:
			if (LIST_FIND(&ip_conntrack_hash[i], 
				      _send_initsync,
			      	       struct ip_conntrack_tuple_hash *,
				       &num_sent)) {
				READ_UNLOCK(&ip_conntrack_lock);
				break;
			}
			if (num_sent < 40 &&
		 	    i+1 < ip_conntrack_htable_size) {
				i++;
				goto dump_bucket_locked;
			}
			READ_UNLOCK(&ip_conntrack_lock);

			CT_SYNC_DEBUG("going to sleep, sent %u\n", num_sent);
			/* schedule until the next bucket */
			__set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ/100);
			__set_current_state(TASK_RUNNING);
			CT_SYNC_DEBUG("woken up\n");
		}
		atomic_set(&ct_sync_initsync_active, 0);

		CT_SYNC_INFO("finished initsync dump\n");
	}

	set_fs(oldmm);

	atomic_set(&stop_ct_initsync_thread, 0);
	wake_up(&stop_ct_initsync_thread_wait);

	CT_SYNC_LEAVE();

	return 0;
}

/* thread startup helper: double fork to reparent to init */
static int
fork_ct_thread(void *p)
{
	struct ct_thread_startup *c = (struct ct_thread_startup *)p;
	CT_SYNC_ENTER();
	
	if (kernel_thread(c->function, c->completion, 0) < 0) {
		CT_SYNC_ERR("Failed to fork ct_sync thread.\n");
		CT_SYNC_LEAVE();
		return -1;
	}

	CT_SYNC_LEAVE();
	
	return 0;
}

/***********************************************************************
 * NOTIFIER HANDLERS
 ***********************************************************************/

/* conntrack notify 'create' function */
static void 
ct_sync_create(struct ip_conntrack *ct)
{
	struct cts_buff *csb;

	CT_SYNC_ENTER();

	if (cts_proto_is_master(cts_cfg.protoh) 
		   && is_confirmed(ct)
#ifdef CONFIG_IP_NF_CONNTRACK_MARK
		   /* generate a message only if the conntrack is synchronized */
		   && test_bit(cmarkbit, &ct->mark)
#endif
	) {
		unsigned char *buff;
		/* This API sucks. you get a cts_buff as return parameter
		 * and have to decrement it after finishing to use it */
		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb, 
				CTMSG_SIZEOF(struct ct_sync_conntrack));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to store create event\n");
			CT_SYNC_LEAVE();
			return;
		}
		fill_ctmsg(buff, CT_SYNC_MSG_UPDATE, ct);
		csb_use_dec(cts_cfg.protoh, csb);
	}

	CT_SYNC_LEAVE();
}

/* conntrack notify 'destroy' function */
static void
ct_sync_destroy(struct ip_conntrack *ct)
{
	struct cts_buff *csb;

	CT_SYNC_ENTER();

	if (cts_proto_is_master(cts_cfg.protoh) 
		   && is_confirmed(ct)
#ifdef CONFIG_IP_NF_CONNTRACK_MARK
		   /* generate a message only if the conntrack is synchronized */
		   && test_bit(cmarkbit, &ct->mark)
#endif
	) {
		unsigned char *buff;
		struct ct_sync_msghdr *hdr;
		struct ip_conntrack_tuple *t;
		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb,
				CTMSG_SIZEOF(*t));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to store destroy event\n");
			CT_SYNC_LEAVE();
			return;
		}
		hdr = (struct ct_sync_msghdr *) buff;
		t = (void *)hdr + sizeof(*hdr);

		/* fill event header */
		hdr->type = CT_SYNC_MSG_DELETE;
		hdr->resource = CT_SYNC_RES_CONNTRACK;
		hdr->len = __constant_htons(sizeof(*t));

		memcpy(t, &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple, sizeof(*t));

		csb_use_dec(cts_cfg.protoh, csb);
	}

	CT_SYNC_LEAVE();
}

/* conntrack expectation created notification */
static void
ct_sync_expect_create(struct ip_conntrack_expect *exp)
{
	struct cts_buff *csb;
	struct ip_conntrack *master_ct = exp->expectant;

	CT_SYNC_ENTER();

	if (likely(cts_proto_is_master(cts_cfg.protoh) &&
	           is_confirmed(master_ct))) {
		void *buff;

		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb,
				CTMSG_SIZEOF(struct ct_sync_expect));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to enqueue event\n");
			CT_SYNC_LEAVE();
			return;
		}
		fill_expectmsg(buff, CT_SYNC_MSG_UPDATE, master_ct, exp);
		csb_use_dec(cts_cfg.protoh, csb);
	}

	CT_SYNC_LEAVE();

	return;
}

/* conntrack expectation destroyed notification */
static void
ct_sync_expect_destroy(struct ip_conntrack_expect *exp)
{
	struct cts_buff *csb;
	struct ip_conntrack *master_ct = exp->expectant;

	CT_SYNC_ENTER();

	if (likely(cts_proto_is_master(cts_cfg.protoh) &&
	    	   is_confirmed(master_ct))) {
		void *buff;

		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb, 
				CTMSG_SIZEOF(struct ct_sync_expect));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to enqueue event\n");
			CT_SYNC_LEAVE();
			return;
		}
		// FIXME: implementation
		csb_use_dec(cts_cfg.protoh, csb);
	}

	CT_SYNC_LEAVE();

	return;
}

/* event processing macros */
#define EVENT(e) (1 << (e))

static int
ct_sync_notify(struct notifier_block *this,
	       unsigned long events, void *ct)
{
	//CT_SYNC_ENTER();

	if ((struct ip_conntrack *)ct == &ip_conntrack_untracked)
		return NOTIFY_DONE;

	if (events & (EVENT(IPCT_NEW) | EVENT(IPCT_STATUS) | EVENT(IPCT_PROTOINFO) |
		      EVENT(IPCT_HELPINFO) | EVENT(IPCT_NATINFO))) {
		CT_SYNC_DEBUG("event: %ld\n", events);
		ct_sync_create((struct ip_conntrack *)ct);
	}
	else if (events & EVENT(IPCT_DESTROY))
		ct_sync_destroy((struct ip_conntrack *)ct);

	//CT_SYNC_LEAVE();

	return NOTIFY_DONE;
}

static struct notifier_block ct_sync_notifier = {
	ct_sync_notify,
	NULL,
	0
};

/***********************************************************************
 * conntrack exemptions
 ***********************************************************************/

static unsigned int
notrack_hook(unsigned int hook,
	     struct sk_buff **pskb,
	     const struct net_device *indev,
	     const struct net_device *outdev,
	     int (*okfn)(struct sk_buff *))
{
	if ((*pskb)->nfct != NULL)
		return NF_ACCEPT;

	/* all traffic coming in or going out sync interface is not
	   to be tracked. Also, loopback traffic is ignored */

	/* FIXME: implement this based on ifindex matching */
	if ((hook == NF_IP_PRE_ROUTING && indev && 
	     indev->ifindex == cts_cfg.ifindex) ||
	    (hook == NF_IP_LOCAL_OUT && outdev &&
	     outdev->ifindex == cts_cfg.ifindex) ||
	    (indev && indev == &loopback_dev) || 
	    (outdev && outdev == &loopback_dev)) {

		/* Attach fake conntrack entry */
		(*pskb)->nfct = &ip_conntrack_untracked.infos[IP_CT_NEW];
		nf_conntrack_get((*pskb)->nfct);
	}

	return NF_ACCEPT;
}

static unsigned int
l2drop_hook(unsigned int hook,
	    struct sk_buff **pskb,
	    const struct net_device *indev,
	    const struct net_device *outdev,
	    int (*okfn)(struct sk_buff *))
{
	/* Allow local traffic */
	if ((indev && indev == &loopback_dev) ||
	    (outdev && outdev == &loopback_dev))
		return NF_ACCEPT;

	/* in slave mode, all traffic _NOT_ on sync interface has to
	   be dropped */
	if (cts_proto_is_master(cts_cfg.protoh))
		return NF_ACCEPT;

	if ((hook == NF_PACKET_OUTPUT && outdev &&
	     outdev->ifindex != cts_cfg.ifindex) ||
	    (hook == NF_PACKET_INPUT && indev &&
	     indev->ifindex != cts_cfg.ifindex)) {
		return NF_DROP;
	}

	return NF_ACCEPT;
}


static struct nf_hook_ops cts_hook_ops[] = {
	{ .hook = notrack_hook,
	  .pf = PF_INET,
	  .hooknum = NF_IP_PRE_ROUTING,
	  .priority = NF_IP_PRI_RAW-1 },
	{ .hook = notrack_hook,
	  .pf = PF_INET,
	  .hooknum = NF_IP_LOCAL_OUT,
	  .priority = NF_IP_PRI_RAW-1 },
	{ .hook = l2drop_hook,
	  .pf = PF_PACKET,
	  .hooknum = NF_PACKET_INPUT,
	  .priority = 0 },
	{ .hook = l2drop_hook,
	  .pf = PF_PACKET,
	  .hooknum = NF_PACKET_OUTPUT,
	  .priority = 0 },
};

/***********************************************************************
 * SYSCTL HANDLING
 ***********************************************************************/

/* sysctl variables */
static int cts_sysctl_state;

static int
cts_sysctl_change_state(int state)
{
	CT_SYNC_ENTER();

	if ((state == 0) && cts_proto_is_master(cts_cfg.protoh)) {
		CT_SYNC_ERR("cannot switch active master to client!\n");
		CT_SYNC_LEAVE();
		return -EINVAL;
	} else if (state == 1) {
		if (!cts_proto_is_master(cts_cfg.protoh)) {

			ct_sync_start_timers();

			if (ip_conntrack_notify_register(&ct_sync_notifier) < 0) {
				CT_SYNC_ERR("cannot register notifier\n");
				CT_SYNC_LEAVE();
				return -EINVAL;
			}

			if (cts_proto_become_master(cts_cfg.protoh) < 0) {
				CT_SYNC_ERR("proto said we cannot become master\n");
				CT_SYNC_LEAVE();
				return -EINVAL;
			}
		} else {
			CT_SYNC_INFO("we already are master; staying cool");
		}
	}

	CT_SYNC_LEAVE();

	return 0;
}

static void
cts_sysctl_update_state(void)
{
	cts_sysctl_state = cts_proto_is_master(cts_cfg.protoh);
}

static int
cts_sysctl_state_handler(ctl_table *table, int write,
		 struct file *filp, void *buffer,
		 size_t *lenp)
{
	int res;
	
	if (write) {
		res = proc_dointvec(table, write, filp, buffer, lenp);
		if (res < 0)
			return res;
		return cts_sysctl_change_state(cts_sysctl_state);
	} else {
		cts_sysctl_update_state();
		return proc_dointvec(table, write, filp, buffer, lenp);
	}
}

/* FIXME: define this in sysctl.h */
#define NET_IPV4_NF_CONNTRACK_SYNC 20

enum {
	CTS_SYSCTL_STATE = 1,
};

static ctl_table ct_sync_sysctl_table[] = {
	{
		.ctl_name	= CTS_SYSCTL_STATE,
		.procname	= "state",
		.mode		= 0644,
		.data		= &cts_sysctl_state,
		.maxlen		= sizeof(int),
		.proc_handler	= &cts_sysctl_state_handler
	},
	{ .ctl_name = 0 }
};

static ctl_table ct_sync_cts_table[] = {
	{
		.ctl_name	= NET_IPV4_NF_CONNTRACK_SYNC,
		.procname	= "ct_sync",
		.mode		= 0555,
		.child		= ct_sync_sysctl_table,
	},
	{ .ctl_name = 0 }
};

static ctl_table ct_sync_netfilter_table[] = {
	{
		.ctl_name	= NET_IPV4_NETFILTER,
		.procname	= "netfilter",
		.mode		= 0555,
		.child		= ct_sync_cts_table,
	},
	{ .ctl_name = 0 }
};

static ctl_table ct_sync_net_table[] = {
	{
		.ctl_name	= NET_IPV4,
		.procname	= "ipv4",
		.mode		= 0555,
		.child		= ct_sync_netfilter_table,
	},
	{ .ctl_name = 0 }
};

static ctl_table ct_sync_root_table[] = {
	{
		.ctl_name	= CTL_NET,
		.procname	= "net",
		.mode		= 0555,
		.child		= ct_sync_net_table,
	},
	{ .ctl_name = 0 }
};

static struct ctl_table_header *ct_sync_sysctl_header;

/***********************************************************************
 * CT_SYNC_PROTO CALLBACK
 ***********************************************************************/

/* This could be converted to notifiers ... */

static void
ct_sync_became_slave(struct cts_protoh *cph, void *data)
{
	ip_conntrack_notify_unregister(&ct_sync_notifier);
	ct_sync_stop_timers();
}

/***********************************************************************
 * MODULE INITIALIZATION
 ***********************************************************************/

/* DO NOT declare this as __init!! */
int init_or_cleanup(int fini)
{
	int ret = -1;
	pid_t pid;
	struct ct_thread_startup startup;
	DECLARE_WAITQUEUE(wait, current);
	DECLARE_COMPLETION(ct_thread_comp);
	struct net_device *sync_netdev;

	if (fini)
		goto cleanup;

	/* informational messages */
	CT_SYNC_DEBUG("ip_ct_tuple: %d bytes\n", sizeof(struct ip_conntrack_tuple));
	CT_SYNC_DEBUG("ct_sync_conntrack: %d bytes\n", sizeof(struct ct_sync_conntrack));
	CT_SYNC_DEBUG("ct_sync_expect: %d bytes\n", sizeof(struct ct_sync_expect));
	CT_SYNC_DEBUG("ct_sync_pkthdr: %d bytes\n", sizeof(struct ct_sync_pkthdr));
	CT_SYNC_DEBUG("ct_sync_msghdr: %d bytes\n", sizeof(struct ct_sync_msghdr));
	CT_SYNC_DEBUG("cmarkbit: %d\n", cmarkbit);
	if (id == -1) {
		CT_SYNC_ERR("Required parameter id is missing.\n");
		return -1;
	}
	if (strcmp(syncdev, "") == 0) {
		CT_SYNC_ERR("Required parameter syncdev is missing.\n");
		return -1;
	}
	if (state == -1) {
		CT_SYNC_ERR("Required parameter state is missing; give 1 for master and 0 for slave.\n");
		return -1;
	}
	if (state != 0 && state != 1) {
		CT_SYNC_ERR("State has to be either 0 or 1\n");
		return -1;
	}

	cts_cfg.nodeid = id;
	strcpy(cts_cfg.devname, syncdev);
	cts_cfg.addr.sin_family = AF_INET;
	cts_cfg.addr.sin_port = __constant_htons(1999);
	cts_cfg.addr.sin_addr.s_addr = __constant_htonl(0xe0000052); /* 224.0.0.82 */

	/* Get ifindex of the sync device, used by the l2drop hook */
	if ((sync_netdev = dev_get_by_name(cts_cfg.devname)) != NULL) {
		cts_cfg.ifindex = sync_netdev->ifindex;
		dev_put(sync_netdev);
	} else {
		CT_SYNC_ERR("Lookup for network interface '%s' failed\n", cts_cfg.devname);
		return -1;
	}

	/* init protocol layer */
	cts_cfg.protoh = cts_proto_init(cts_cfg.devname, &cts_cfg.addr, id,
					state, &ct_sync_became_slave, NULL,
					&ct_sync_send_wait);
	if (!cts_cfg.protoh) {
		CT_SYNC_ERR("Failed to initialize protocol.\n");
		ret = -EINVAL;
		goto error_none;
	}

	/* Register hooks first, make sure not even the first sync packet gets
	 * tracked */
	if (nf_register_hook(&cts_hook_ops[0]) < 0)
		goto error_proto;
	if (nf_register_hook(&cts_hook_ops[1]) < 0)
		goto error_hook0;
	if (l2drop) {
		if (nf_register_hook(&cts_hook_ops[2]) < 0)
			goto error_hook1;
		if (nf_register_hook(&cts_hook_ops[3]) < 0)
			goto error_hook2;
		ip_ct_selective_cleanup(kill_all, NULL);
	}

	/* register sysctl entries */
	ct_sync_sysctl_header = register_sysctl_table(ct_sync_root_table, 0);
	if (ct_sync_sysctl_header == NULL)
		goto error_hook3;

	/* start threads */
	startup.completion = &ct_thread_comp;
	startup.function = ct_sync_send_thread_main;
	if ((pid = kernel_thread(fork_ct_thread, &startup, 0)) < 0)
	{
		CT_SYNC_ERR("Failed to start fork thread.\n");
		ret = pid;
		goto error_proc;
	}
	if ((ret = waitpid(pid, NULL, __WCLONE)) != pid)
	{
		CT_SYNC_ERR("Fork thread exit problem.\n");
		goto error_proc;
	}
	wait_for_completion(&ct_thread_comp);

	init_completion(&ct_thread_comp);
	startup.completion = &ct_thread_comp;
	startup.function = ct_sync_rcv_thread_main;
	if ((pid = kernel_thread(fork_ct_thread, &startup, 0)) < 0)
	{
		CT_SYNC_ERR("Failed to start fork thread.\n");
		ret = pid;
		goto error_sthread;
	}
	if ((ret = waitpid(pid, NULL, __WCLONE)) != pid)
	{
		CT_SYNC_ERR("Fork thread exit problem.\n");
		goto error_sthread;
	}
	wait_for_completion(&ct_thread_comp);

	init_completion(&ct_thread_comp);
	startup.completion = &ct_thread_comp;
	startup.function = ct_sync_initsync_thread_main;
	if ((pid = kernel_thread(fork_ct_thread, &startup, 0)) < 0)
	{
		CT_SYNC_ERR("Failed to start fork thread.\n");
		ret = pid;
		goto error_rthread;
	}
	if ((ret = waitpid(pid, NULL, __WCLONE)) != pid)
	{
		CT_SYNC_ERR("Fork thread exit problem.\n");
		goto error_rthread;
	}
	wait_for_completion(&ct_thread_comp);

	/* Now we have all of our infrastructure up and running */

	if (state) {
		/* master: register notify callbacks */
		ret = ip_conntrack_notify_register(&ct_sync_notifier);
		if (ret < 0) {
			CT_SYNC_ERR("unable to register notifier\n");
			goto error_ithread;
		}

		if (cts_proto_become_master(cts_cfg.protoh) < 0) {
			ret = -1;
			goto cleanup;
		}
	} else {
		/* slave: stop timers of connections marked for synchronization
		 * and request initsync */
		ct_sync_stop_timers();

		if (cts_proto_request_initsync(cts_cfg.protoh) < 0) {
			ret = -1;
			goto error_ithread;
		}
	}

	printk(KERN_NOTICE "netfilter conntrack_sync Version %s loaded\n",
		CT_SYNC_VERSION);
	CT_SYNC_INFO("parameters: 'id=%u state=%u syncdev=%s l2drop=%u'\n",
			id, state, syncdev, l2drop);

        return 0;

cleanup:
	CT_SYNC_DEBUG("unregistering conntrack notifier\n");
	ip_conntrack_notify_unregister(&ct_sync_notifier);
error_ithread:
	CT_SYNC_DEBUG("stopping initsync thread\n");
	__set_current_state(TASK_INTERRUPTIBLE);
	wake_up(&ct_sync_initsync_wait);
	atomic_set(&stop_ct_initsync_thread, 1);
	add_wait_queue(&stop_ct_initsync_thread_wait, &wait);
	schedule();
	__set_current_state(TASK_RUNNING);
	remove_wait_queue(&stop_ct_initsync_thread_wait, &wait);
error_rthread:
	CT_SYNC_DEBUG("stopping receive thread\n");
	__set_current_state(TASK_INTERRUPTIBLE);
	atomic_set(&stop_ct_rcv_thread, 1);
	add_wait_queue(&stop_ct_rcv_thread_wait, &wait);
	schedule();
	__set_current_state(TASK_RUNNING);
	remove_wait_queue(&stop_ct_rcv_thread_wait, &wait);
error_sthread:
	CT_SYNC_DEBUG("stopping send thread\n");
	__set_current_state(TASK_INTERRUPTIBLE);
	atomic_set(&stop_ct_send_thread, 1);
	add_wait_queue(&stop_ct_send_thread_wait, &wait);
	wake_up(&ct_sync_send_wait);
	schedule();
	__set_current_state(TASK_RUNNING);
	remove_wait_queue(&stop_ct_send_thread_wait, &wait);
error_proc:
	CT_SYNC_DEBUG("removing sysctl entry\n");
	unregister_sysctl_table(ct_sync_sysctl_header);
error_hook3:
	if (l2drop)
		nf_unregister_hook(&cts_hook_ops[3]);
error_hook2:
	if (l2drop)
		nf_unregister_hook(&cts_hook_ops[2]);
error_hook1:
	nf_unregister_hook(&cts_hook_ops[1]);
error_hook0:
	nf_unregister_hook(&cts_hook_ops[0]);
error_proto:
	cts_proto_cleanup(cts_cfg.protoh);
	/* make sure all conntrack timers are running */
	ct_sync_start_timers();
error_none:
	CT_SYNC_DEBUG("returning %d\n", ret);
	return ret;
}

int __init init(void)
{
	return init_or_cleanup(0);
}

/* remove the module */
void __exit fini(void)
{
	init_or_cleanup(1);
	printk(KERN_NOTICE "netfilter conntrack_sync Version %s unloaded\n",
		CT_SYNC_VERSION);
}

module_init(init);
module_exit(fini);

